Checklist = class(Turbine.UI.Lotro.Window);

local traitPointIcon = Turbine.UI.Graphic(0x41142C2E);
local expandIcon = Turbine.UI.Graphic(0x41007E27);
local collapseIcon = Turbine.UI.Graphic(0x41007E26);
local valarIcon = Turbine.UI.Graphic(0x41004285);
local player = Turbine.Gameplay.LocalPlayer:GetInstance();
local playerLevel = player:GetLevel();
local playerClassId = player:GetClass();
local maxLevel = 140;

function Checklist:Constructor(settings, oldOrderedCheckBoxes, oldOrderedContainers)
    Turbine.UI.Lotro.Window.Constructor(self);

    -- Default settings
    self.settings = {
        language = settings and (settings.language) or Turbine.Engine.GetLanguage();
        size = { 670, 600 };
        visible = true;
        checkBoxes = {};
        expanded = {};
        icon = {
            position = { 0.5, 0.5 }
        }
    };

    L:SetContext("/ChecklistWindow");
    self:SetText(L:GetText("Title"));
    Turbine.UI.Lotro.Window.SetVisible(self, true);
    self:SetResizable(true);
    self:SetMinimumWidth(670);
	AddCallback(self, "SizeChanged", function() self:_SizeChanged() end);

    self.loading = true;
    self.checkBoxes = {};
    self.oldOrderedCheckBoxes = oldOrderedCheckBoxes;
    self.orderedCheckBoxes = {};    
    self.containers = {};
    self.oldOrderedContainers = oldOrderedContainers;
    self.orderedContainers = {};
    self.haveAria = false;

    self.icon = Thurallor.UI.Icon(nil, self, self.settings.icon);
    for _, pos in ipairs({ {-5, -4}, {3, 3} }) do
        local img = Turbine.UI.Control();
        img:SetBackground(traitPointIcon);
        img:SetSize(32, 32);
        img:SetParent(self.icon);
        img:SetBlendMode(Turbine.UI.BlendMode.AlphaBlend);
        img:SetPosition(unpack(pos));
        img:SetMouseVisible(false);
    end
    Thurallor.UI.Tooltip(self:GetText()):Attach(self.icon);
    AddCallback(self.icon, "SettingsChanged", function()
        self:SaveSettings();
    end);

    self:Redraw();
    if (settings) then
        self:ApplySettings(settings);
        self.loading = false;
    else
        Turbine.PluginData.Load(Turbine.DataScope.Character, "ClassTraitPoints", function(data)
            if (data) then
                data = ImportTable(data);
                self:ApplySettings(data);
            else
                self:ApplySettings(self.settings);
            end
            self.loading = false;
        end);
    end

    -- Redraw the list when player level increases
    local player = Turbine.Gameplay.LocalPlayer:GetInstance();
    AddCallback(player, "LevelChanged", function()
        self:Redraw(player:GetLevel());
    end);
end

function Checklist:ApplySettings(data)
    data.haveAria = nil; -- obsolete
    data.language = data.language or Turbine.Engine.GetLanguage();
    if (data.icon) then
        self.icon:ApplySettings(data.icon);
        Turbine.UI.Lotro.Window.SetVisible(self, data.visible);
    else
        data.icon = { position = { 0.5, 0.5 } };
        data.visible = true;
    end
    self:SetSize(unpack(data.size));
    if (data.position) then
        self:SetPosition(unpack(data.position));
    else
        CenterWindow(self);
    end
    if (self.oldOrderedCheckBoxes) then
        -- Restore checkboxes to the same state they were in before the language change
        for n = 1, #self.oldOrderedCheckBoxes do
            local checkBox = self.oldOrderedCheckBoxes[n];
            local checked = checkBox:IsChecked();
            self.orderedCheckBoxes[n]:SetChecked(checked);
        end
        self.oldOrderedCheckBoxes = nil;
    else
        -- Restore checkboxes to the saved state
        for text, value in pairs(data.checkBoxes) do
            local checkBox = self.checkBoxes[text];
            if (checkBox) then
                checkBox:SetChecked(value);
            end
        end
    end
    data.expanded = data.expanded or {};
    if (self.oldOrderedContainers) then
        -- Restore containers to the same state they were in before the language change
        for n = 1, #self.oldOrderedContainers do
            local container = self.oldOrderedContainers[n];
            local expanded = container:IsExpanded();
            self.orderedContainers[n]:SetExpanded(expanded);
        end
        self.oldOrderedContainers = nil;
    else
        -- Restore containers to the saved state
        for text, value in pairs(data.expanded) do
            local container = self.containers[text];
            if (container) then
                container:SetExpanded(value);
            end
        end
    end
    self:RecalculateTotal();
    self.settings = data;
    if (data.language ~= L:GetLanguage()) then
        return self:ChangeLanguage(data.language);
    end
end

function Checklist:Redraw(level, classId)
    L:SetContext("/ChecklistWindow");
    playerLevel = level or playerLevel;
    playerClassId = classId or playerClassId;
    local classId, playerClassStr = {};
    for name, id in pairs(Turbine.Gameplay.Class) do
        local classStr = L:GetText("PlayerClass/" .. name);
        classId[classStr] = id;
        if (playerClassId == id) then
            playerClass = name;
            playerClassStr = classStr;
        end
    end
    local font = Turbine.UI.Lotro.Font.Verdana16;
    local fontHeight = 16;
    local label, checkBox, text, help;

    self:GetControls():Clear();
    self.checkBoxes = {};
    self.orderedCheckBoxes = {};
    self.containers = {};
    self.orderedContainers = {};

    -- Display pulldown menus for class and level
    self.settingsCtl = Turbine.UI.Control();
    self.settingsCtl:SetParent(self);
    self.settingsCtl:SetTop(30);
    self.settingsCtl:SetHeight(20);
    AddCallback(self, "SizeChanged", function()
        self.settingsCtl:SetLeft(math.floor(0.5 + (self:GetWidth() - self.settingsCtl:GetWidth()) / 2));
    end);

    -- Display pulldown menu for level
    local levels = {};
    for level = 1, maxLevel do
        local levelStr = L:GetText("PlayerLevel"):gsub("<level>", level);
        table.insert(levels, levelStr);
    end
    local playerLevelStr = L:GetText("PlayerLevel"):gsub("<level>", playerLevel);
    self.levelSelector = Thurallor.UI.DropDown(levels, playerLevelStr);
    self.levelSelector:SetParent(self.settingsCtl);
    self.levelSelector:SetWidth(95);
    AddCallback(self.levelSelector, "ItemChanged", function(_, args)
        local level = args.Index;
        self:Redraw(level, nil);
    end);

    -- Display pulldown menu for class
    local classes = L:GetSortedTexts("PlayerClass");
    self.classSelector = Thurallor.UI.DropDown(classes, playerClassStr);
    self.classSelector:SetParent(self.settingsCtl);
    self.classSelector:SetWidth(145);
    self.classSelector:SetLeft(97);
    AddCallback(self.classSelector, "ItemChanged", function(_, args)
        local classStr = args.Text;
        self:Redraw(nil, classId[classStr]);
    end);
    self.settingsCtl:SetWidth(97 + 145);

    -- Display pulldown menu for language
    local languages = { "en", "de", "fr", "ru" };
    local langMap = {
        [Turbine.Language.English] = "en";
        [Turbine.Language.German] = "de";
        [Turbine.Language.French] = "fr";
        [Turbine.Language.Russian] = "ru";
    };
    self.languageSelector = Thurallor.UI.DropDown(languages, langMap[self.settings.language]);
    self.languageSelector:SetParent(self);
    self.languageSelector:SetWidth(45);
    self.languageSelector:SetTop(30);
    AddCallback(self, "SizeChanged", function()
        self.languageSelector:SetLeft(self:GetWidth() - self.languageSelector:GetWidth() - 30);
    end);
    AddCallback(self.languageSelector, "ItemChanged", function(_, args)
        local langMap = {
            en = Turbine.Language.English;
            de = Turbine.Language.German;
            fr = Turbine.Language.French;
            ru = Turbine.Language.Russian;
        };
        self:ChangeLanguage(langMap[args.Text]);
    end);
    
    -- Display total and maximum points
    self.total = Turbine.UI.Label();
    self.total:SetParent(self);
    self.total:SetMouseVisible(false);
    self.total:SetForeColor(Turbine.UI.Color(1, 0, 1, 0));
    self.total:SetFont(Turbine.UI.Lotro.Font.TrajanPro24);
    self.total:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter);
    self.total:SetHeight(40);
    self.total:SetTop(52);
    AddCallback(self, "SizeChanged", function()
        self.total:SetWidth(self:GetWidth());
    end);

    local leftMargin, topMargin, rightMargin, bottomMargin = 20, 90, 20, 24;
    self.hScrollBar = Turbine.UI.Lotro.ScrollBar();
    self.hScrollBar:SetOrientation(Turbine.UI.Orientation.Horizontal);
    self.hScrollBar:SetParent(self);
    self.vScrollBar = Turbine.UI.Lotro.ScrollBar();
    self.vScrollBar:SetOrientation(Turbine.UI.Orientation.Vertical);
    self.vScrollBar:SetParent(self);
    self.interior = Turbine.UI.ListBox();
    self.interior:SetParent(self);
    self.interior:SetPosition(leftMargin, topMargin);
    self.interior:SetHorizontalScrollBar(self.hScrollBar);
    self.interior:SetVerticalScrollBar(self.vScrollBar);
    AddCallback(self, "SizeChanged", function()
        local width, height = self:GetSize();
        self.interior:SetSize(width - leftMargin - rightMargin, height - topMargin - bottomMargin);
        self.hScrollBar:SetSize(width - 60, 10);
        self.hScrollBar:SetPosition(30, height - 16);
        self.vScrollBar:SetSize(10, height - 76);
        self.vScrollBar:SetPosition(width - 14, 46);
    end);

    local function AddLabel(text)
        local label = Turbine.UI.Label();
        label:SetFont(font);
        label:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
        label:SetMarkupEnabled(true);
        label:SetText(text);
        self.interior:AddItem(label);
        label:SetHeight(40);
        AddCallback(self, "SizeChanged", function()
            label:SetWidth(self.interior:GetWidth());
        end);
        return label;
    end

    local function AddItem(text, help, ariaHides)
        local checked = self.settings.checkBoxes[text];
        local container = Turbine.UI.Control();
        container:SetHeight(18);

        -- Add a checkbox
        local checkBox = Turbine.UI.Lotro.CheckBox();
        checkBox:SetChecked(checked);
        checkBox:SetHeight(16);
        checkBox:SetFont(font);
        checkBox:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
        checkBox:SetText("    " .. text);
        checkBox:SetMarkupEnabled(true);
        checkBox.text = text;
        checkBox:SetParent(container);
        AddCallback(checkBox, "CheckedChanged", function(obj)
            self.settings.checkBoxes[obj.text] = obj:IsChecked();
            self:RecalculateTotal();
        end);
        container.checkBox = checkBox;

        -- Add an "expand/collapse" button
        local expandCollapse = Turbine.UI.Control();
        expandCollapse:SetSize(16, 16);
        expandCollapse:SetParent(container);
        expandCollapse:SetLeft(20);
        expandCollapse:SetBackground(expandIcon);
        expandCollapse:SetBlendMode(Turbine.UI.BlendMode.Overlay);
        expandCollapse.MouseClick = function(obj)
            local container = obj:GetParent();
            container:SetExpanded(not container:IsExpanded());
        end
        container.expandCollapse = expandCollapse;

        -- When you click on the text, we want it to expand/collapse rather than check/uncheck
        local textMask = Turbine.UI.Control();
        textMask:SetHeight(16);
        textMask:SetParent(container);
        textMask:SetLeft(36);
        textMask.MouseClick = function(_, args)
            expandCollapse:MouseClick(args);
        end

        -- Add the help text when expanded
        help:SetParent(container);
        help:SetPosition(20, 18);
        container.help = help;

        function container.IsExpanded(cont)
            return (cont:GetHeight() > 18);
        end
        function container.SetExpanded(cont, expanded)
            local fullHeight = 18 + cont.help:GetHeight();
            if (expanded) then
                cont:SetHeight(fullHeight);
                cont.expandCollapse:SetBackground(collapseIcon);
            else
                cont:SetHeight(18);
                cont.expandCollapse:SetBackground(expandIcon);
            end
            self.settings.expanded[cont.checkBox.text] = expanded;
            self:SaveSettings();
        end
        container:SetExpanded(self.settings.expanded[text]);

        self.checkBoxes[text] = checkBox;
        table.insert(self.orderedCheckBoxes, checkBox);
        self.containers[text] = container;
        table.insert(self.orderedContainers, container);
        self.settings.checkBoxes[text] = checked;
        if (ariaHides and self.haveAria) then
            checkBox.visible = false;
        else
            checkBox.visible = true;
            self.interior:AddItem(container);
            AddCallback(self, "SizeChanged", function()
                local width = self.interior:GetWidth();
                container:SetWidth(width);
                checkBox:SetWidth(width);
                textMask:SetWidth(width);
                help:SetWidth(width - 20);
            end);
        end
        return checkBox;
    end
    
    local function CreateHelp(text, url, maxLines)
        local help = Turbine.UI.Label();
        help:SetSelectable(true);
        help:SetMultiline(true);
        help:SetMarkupEnabled(true);
        help:SetFont(font);
        help:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
        text = "<rgb=#C0C0C0>" .. text .. "</rgb>";
        if (url) then
            text = text .. "\n<rgb=#00C0C0>https://lotro-wiki.com/index.php/" .. url .. "</rgb>";
        end
        help:SetText(text);
        help:SetHeight((maxLines + 0.5) * fontHeight);
        return help;
    end

    local text = L:GetText("Description"):gsub("<class>", playerClassStr):gsub("<level>", playerLevel);
    label = AddLabel(text);
    label:SetHeight((L:GetText("DescriptionHeight") + 0.5) * fontHeight);
    L:SetContext("Items");

    -- Aria of the Valar
    L:SetContext("/ChecklistWindow/Items/AriaOfTheValar");
    self.ariaPoints = 0;
    self.ariaCheckBox = nil;
    self.haveAria = false;
    if (playerLevel >= 105) then
        if ((playerClass == "Beorning") or (playerClass == "Brawler")) then
            self.ariaPoints = 72;
        else
            self.ariaPoints = 75;
        end
        text = L:GetText("Help"):gsub("<class>", playerClassStr):gsub("<points>", self.ariaPoints - 61);
        help = CreateHelp(text, "Aria_of_the_Valar_Trait_Points", L:GetText("Height"));
        text = L:GetText("Desc"):gsub("<points>", self.ariaPoints);
        checkBox = AddItem(text, help, false);
        self.haveAria = checkBox:IsChecked();
        AddCallback(checkBox, "CheckedChanged", function(ctl)
            self.haveAria = ctl:IsChecked();
            self:Redraw();
        end);
        self.ariaCheckBox = checkBox;
    end

    -- Points granted by player level
    L:SetContext("/ChecklistWindow/Items/ForReachingLevel");
    for _, level in pairs({
        7, 9,
        11, 13, 15, 17, 19,
        21, 23, 25, 27, 29,
        31, 33, 35, 37, 39,
        41, 43, 45, 47, 49,
        51, 53, 55, 57, 59,
        61, 63, 65, 67, 69,
        71, 73, 75, 77, 79,
        81, 83, 85, 87, 89,
        91, 93, 95, 97, 99,
        101, 103, 105, 108,
        111, 114, 120, 140
    }) do
        if (level <= playerLevel) then
            local ariaHides = (level <= 105);
            text = L:GetText("Help"):gsub("<level>", level);
            help = CreateHelp(text, nil, L:GetText("Height"));
            text = L:GetText("Desc"):gsub("<level>", level);
            if (self.settings.checkBoxes[text] == nil) then
                self.settings.checkBoxes[text] = true;
            end
            checkBox = AddItem(text, help, ariaHides);
        end
    end

    -- Points granted from class deeds
    L:SetContext("/ChecklistWindow/Items/ClassDeeds");
    local classDeeds = {
        Beorning = {
            { name = L:GetText("Tiered"):gsub("<tier>", 1), minLevel = 5 , url = "Class_Deeds_(Beorning)_-_Tier_1"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 2), minLevel = 13, url = "Class_Deeds_(Beorning)_-_Tier_2"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 3), minLevel = 16, url = "Class_Deeds_(Beorning)_-_Tier_3"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 4), minLevel = 20, url = "Class_Deeds_(Beorning)_-_Tier_4"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 5), minLevel = 30, url = "Class_Deeds_(Beorning)_-_Tier_5"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 6), minLevel = 40, url = "Class_Deeds_(Beorning)_-_Tier_6"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 7), minLevel = 44, url = "Class_Deeds_(Beorning)_-_Tier_7"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 8), minLevel = 58, url = "Class_Deeds_(Beorning)_-_Tier_8"            , section = "Class" };
            { name = L:GetText("Beorning1")               , minLevel = 39, url = "A Hobbit's Holiday (Beorning Deed)"         , section = "Class" };
            { name = L:GetText("Beorning2")               , minLevel = 39, url = "A_Study_of_the_Skin-changer_(Beorning_Deed)", section = "Class" };
            { name = L:GetText("Beorning3")               , minLevel = 39, url = "Geneology_of_the_Beornings_(Beorning_Deed)" , section = "Class" };
        };
        Brawler = {
            { name = L:GetText("Tiered"):gsub("<tier>", 1), minLevel = 10, url = "Class_Deeds_(Brawler)_-_Tier_1"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 2), minLevel = 10, url = "Class_Deeds_(Brawler)_-_Tier_2"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 3), minLevel = 15, url = "Class_Deeds_(Brawler)_-_Tier_3"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 4), minLevel = 24, url = "Class_Deeds_(Brawler)_-_Tier_4"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 5), minLevel = 30, url = "Class_Deeds_(Brawler)_-_Tier_5"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 6), minLevel = 40, url = "Class_Deeds_(Brawler)_-_Tier_6"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 7), minLevel = 45, url = "Class_Deeds_(Brawler)_-_Tier_7"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 8), minLevel = 50, url = "Class_Deeds_(Brawler)_-_Tier_8"             , section = "Class" };
            { name = L:GetText("Brawler1")                , minLevel = 39, url = "Martial_Movement"                           , section = "Class" };
            { name = L:GetText("Brawler2")                , minLevel = 39, url = "Knowledge_of_Kicks"                         , section = "Class" };
            { name = L:GetText("Brawler3")                , minLevel = 39, url = "The_Art_of_the_Perfect_Punch"               , section = "Class" };
        };
        Burglar = {
            { name = L:GetText("Tiered"):gsub("<tier>", 1), minLevel = 5 , url = "Class_Deeds_(Burglar)_-_Tier_1"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 2), minLevel = 10, url = "Class_Deeds_(Burglar)_-_Tier_2"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 3), minLevel = 15, url = "Class_Deeds_(Burglar)_-_Tier_3"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 4), minLevel = 20, url = "Class_Deeds_(Burglar)_-_Tier_4"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 5), minLevel = 30, url = "Class_Deeds_(Burglar)_-_Tier_5"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 6), minLevel = 30, url = "Class_Deeds_(Burglar)_-_Tier_6"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 7), minLevel = 40, url = "Class_Deeds_(Burglar)_-_Tier_7"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 8), minLevel = 50, url = "Class_Deeds_(Burglar)_-_Tier_8"             , section = "Class" };
            { name = L:GetText("Burglar1")                , minLevel = 39, url = "Knee-breaker%27s_Manual"                    , section = "Class" };
            { name = L:GetText("Burglar2")                , minLevel = 39, url = "The_Book_of_Knives"                         , section = "Class" };
            { name = L:GetText("Burglar3")                , minLevel = 39, url = "The_Expert%27s_Guide_to_Dirty_Fighting"     , section = "Class" };
            { name = L:GetText("MinesOfMoria")            , minLevel = 55, url = "The_Mines_of_Moria_(Trait_Point_Deed)"      , section = "Epic"  };
            { name = L:GetText("Burglar4")                , minLevel = 58, url = "The_Path_of_the_Mischief-maker"             , section = "Class" };
            { name = L:GetText("Burglar5")                , minLevel = 58, url = "A_Guide_to_the_Quiet_Knife"                 , section = "Class" };
        };
        Captain = {
            { name = L:GetText("Tiered"):gsub("<tier>", 1), minLevel = 5 , url = "Class_Deeds_(Captain)_-_Tier_1"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 2), minLevel = 10, url = "Class_Deeds_(Captain)_-_Tier_2"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 3), minLevel = 16, url = "Class_Deeds_(Captain)_-_Tier_3"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 4), minLevel = 20, url = "Class_Deeds_(Captain)_-_Tier_4"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 5), minLevel = 30, url = "Class_Deeds_(Captain)_-_Tier_5"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 6), minLevel = 45, url = "Class_Deeds_(Captain)_-_Tier_6"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 7), minLevel = 50, url = "Class_Deeds_(Captain)_-_Tier_7"             , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 8), minLevel = 50, url = "Class_Deeds_(Captain)_-_Tier_8"             , section = "Class" };
            { name = L:GetText("Captain1")                , minLevel = 39, url = "The_Book_of_Oaths"                          , section = "Class" };
            { name = L:GetText("Captain2")                , minLevel = 39, url = "The_Candle%27s_Flame"                       , section = "Class" };
            { name = L:GetText("Captain3")                , minLevel = 39, url = "The_Treatise_of_Valour"                     , section = "Class" };
            { name = L:GetText("MinesOfMoria")            , minLevel = 55, url = "The_Mines_of_Moria_(Trait_Point_Deed)"      , section = "Epic"  };
            { name = L:GetText("Captain4")                , minLevel = 58, url = "The_Path_of_the_Healing_Hands"              , section = "Class" };
            { name = L:GetText("Captain5")                , minLevel = 58, url = "The_Master_of_the_Charge"                   , section = "Class" };
        };
        Champion = {
            { name = L:GetText("Tiered"):gsub("<tier>", 1), minLevel = 5 , url = "Class_Deeds_(Champion)_-_Tier_1"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 2), minLevel = 10, url = "Class_Deeds_(Champion)_-_Tier_2"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 3), minLevel = 15, url = "Class_Deeds_(Champion)_-_Tier_3"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 4), minLevel = 20, url = "Class_Deeds_(Champion)_-_Tier_4"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 5), minLevel = 30, url = "Class_Deeds_(Champion)_-_Tier_5"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 6), minLevel = 40, url = "Class_Deeds_(Champion)_-_Tier_6"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 7), minLevel = 45, url = "Class_Deeds_(Champion)_-_Tier_7"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 8), minLevel = 50, url = "Class_Deeds_(Champion)_-_Tier_8"            , section = "Class" };
            { name = L:GetText("Champion1")               , minLevel = 39, url = "The_Artisan_Blade"                          , section = "Class" };
            { name = L:GetText("Champion2")               , minLevel = 39, url = "The_Joy_of_Battle"                          , section = "Class" };
            { name = L:GetText("Champion3")               , minLevel = 39, url = "The_Tome_of_Swords"                         , section = "Class" };
            { name = L:GetText("MinesOfMoria")            , minLevel = 55, url = "The_Mines_of_Moria_(Trait_Point_Deed)"      , section = "Epic"  };
            { name = L:GetText("Champion4")               , minLevel = 58, url = "The_Path_of_the_Martial_Champion_(Deed)"    , section = "Class" };
            { name = L:GetText("Champion5")               , minLevel = 58, url = "The_Boiling_Rage_(Deed)"                    , section = "Class" };
        };
        Guardian = {
            { name = L:GetText("Tiered"):gsub("<tier>", 1), minLevel = 26, url = "Class_Deeds_(Guardian)_-_Tier_1"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 2), minLevel = 30, url = "Class_Deeds_(Guardian)_-_Tier_2"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 3), minLevel = 15, url = "Class_Deeds_(Guardian)_-_Tier_3"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 4), minLevel = 20, url = "Class_Deeds_(Guardian)_-_Tier_4"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 5), minLevel = 30, url = "Class_Deeds_(Guardian)_-_Tier_5"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 6), minLevel = 40, url = "Class_Deeds_(Guardian)_-_Tier_6"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 7), minLevel = 40, url = "Class_Deeds_(Guardian)_-_Tier_7"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 8), minLevel = 50, url = "Class_Deeds_(Guardian)_-_Tier_8"            , section = "Class" };
            { name = L:GetText("Guardian1")               , minLevel = 39, url = "A_Shield-maiden%27s_Song"                   , section = "Class" };
            { name = L:GetText("Guardian2")               , minLevel = 39, url = "The_Best_Defence"                           , section = "Class" };
            { name = L:GetText("Guardian3")               , minLevel = 39, url = "The_Final_Word"                             , section = "Class" };
            { name = L:GetText("MinesOfMoria")            , minLevel = 55, url = "The_Mines_of_Moria_(Trait_Point_Deed)"      , section = "Epic"  };
            { name = L:GetText("Guardian4")               , minLevel = 58, url = "The_Path_of_Freedom%27s_Defender"           , section = "Class" };
            { name = L:GetText("Guardian5")               , minLevel = 58, url = "A_Keen_Blade"                               , section = "Class" };
        };
        Hunter = {
            { name = L:GetText("Tiered"):gsub("<tier>", 1), minLevel = 5 , url = "Class_Deeds_(Hunter)_-_Tier_1"              , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 2), minLevel = 10, url = "Class_Deeds_(Hunter)_-_Tier_2"              , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 3), minLevel = 15, url = "Class_Deeds_(Hunter)_-_Tier_3"              , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 4), minLevel = 20, url = "Class_Deeds_(Hunter)_-_Tier_4"              , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 5), minLevel = 38, url = "Class_Deeds_(Hunter)_-_Tier_5"              , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 6), minLevel = 40, url = "Class_Deeds_(Hunter)_-_Tier_6"              , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 7), minLevel = 48, url = "Class_Deeds_(Hunter)_-_Tier_7"              , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 8), minLevel = 50, url = "Class_Deeds_(Hunter)_-_Tier_8"              , section = "Class" };
            { name = L:GetText("Hunter1")                 , minLevel = 39, url = "A_Shot_in_the_Dark"                         , section = "Class" };
            { name = L:GetText("Hunter2")                 , minLevel = 39, url = "The_Furthest_Charge"                        , section = "Class" };
            { name = L:GetText("Hunter3")                 , minLevel = 39, url = "The_Way_of_the_Hunter"                      , section = "Class" };
            { name = L:GetText("MinesOfMoria")            , minLevel = 55, url = "The_Mines_of_Moria_(Trait_Point_Deed)"      , section = "Epic"  };
            { name = L:GetText("Hunter4")                 , minLevel = 58, url = "The_Path_of_the_Foe-trapper"                , section = "Class" };
            { name = L:GetText("Hunter5")                 , minLevel = 58, url = "The_Jolly_Hunter"                           , section = "Class" };
        };
        LoreMaster = {
            { name = L:GetText("Tiered"):gsub("<tier>", 1), minLevel = 16, url = "Class_Deeds_(Lore-master)_-_Tier_1"         , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 2), minLevel = 15, url = "Class_Deeds_(Lore-master)_-_Tier_2"         , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 3), minLevel = 16, url = "Class_Deeds_(Lore-master)_-_Tier_3"         , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 4), minLevel = 20, url = "Class_Deeds_(Lore-master)_-_Tier_4"         , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 5), minLevel = 30, url = "Class_Deeds_(Lore-master)_-_Tier_5"         , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 6), minLevel = 30, url = "Class_Deeds_(Lore-master)_-_Tier_6"         , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 7), minLevel = 40, url = "Class_Deeds_(Lore-master)_-_Tier_7"         , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 8), minLevel = 50, url = "Class_Deeds_(Lore-master)_-_Tier_8"         , section = "Class" };
            { name = L:GetText("LoreMaster1")             , minLevel = 39, url = "Of_Leaf_and_Twig"                           , section = "Class" };
            { name = L:GetText("LoreMaster2")             , minLevel = 39, url = "The_Book_of_Beasts"                         , section = "Class" };
            { name = L:GetText("LoreMaster3")             , minLevel = 39, url = "Lore_of_the_Blade"                          , section = "Class" };
            { name = L:GetText("MinesOfMoria")            , minLevel = 55, url = "The_Mines_of_Moria_(Trait_Point_Deed)"      , section = "Epic"  };
            { name = L:GetText("LoreMaster4")             , minLevel = 58, url = "The_Path_of_the_Ancient_Master"             , section = "Class" };
            { name = L:GetText("LoreMaster5")             , minLevel = 58, url = "The_Book_of_Nature"                         , section = "Class" };
        };
        Minstrel = {
            { name = L:GetText("Tiered"):gsub("<tier>", 1), minLevel = 5 , url = "Class_Deeds_(Minstrel)_-_Tier_1"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 2), minLevel = 10, url = "Class_Deeds_(Minstrel)_-_Tier_2"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 3), minLevel = 15, url = "Class_Deeds_(Minstrel)_-_Tier_3"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 4), minLevel = 20, url = "Class_Deeds_(Minstrel)_-_Tier_4"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 5), minLevel = 30, url = "Class_Deeds_(Minstrel)_-_Tier_5"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 6), minLevel = 30, url = "Class_Deeds_(Minstrel)_-_Tier_6"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 7), minLevel = 45, url = "Class_Deeds_(Minstrel)_-_Tier_7"            , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 8), minLevel = 50, url = "Class_Deeds_(Minstrel)_-_Tier_8"            , section = "Class" };
            { name = L:GetText("Minstrel1")               , minLevel = 39, url = "Melodies_of_the_Valar"                      , section = "Class" };
            { name = L:GetText("Minstrel2")               , minLevel = 39, url = "Valour%27s_Marches"                         , section = "Class" };
            { name = L:GetText("Minstrel3")               , minLevel = 39, url = "Lore_of_the_Blade"                          , section = "Class" };
            { name = L:GetText("MinesOfMoria")            , minLevel = 55, url = "The_Mines_of_Moria_(Trait_Point_Deed)"      , section = "Epic"  };
            { name = L:GetText("Minstrel4")               , minLevel = 58, url = "The_Path_of_the_Resolve-watcher"            , section = "Class" };
            { name = L:GetText("Minstrel5")               , minLevel = 58, url = "The_Verses_of_the_North"                    , section = "Class" };
        };
        RuneKeeper = {
            { name = L:GetText("Tiered"):gsub("<tier>", 1), minLevel = 5 , url = "Class_Deeds_(Rune-keeper)_-_Tier_1"         , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 2), minLevel = 15, url = "Class_Deeds_(Rune-keeper)_-_Tier_2"         , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 3), minLevel = 15, url = "Class_Deeds_(Rune-keeper)_-_Tier_3"         , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 4), minLevel = 20, url = "Class_Deeds_(Rune-keeper)_-_Tier_4"         , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 5), minLevel = 30, url = "Class_Deeds_(Rune-keeper)_-_Tier_5"         , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 6), minLevel = 36, url = "Class_Deeds_(Rune-keeper)_-_Tier_6"         , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 7), minLevel = 40, url = "Class_Deeds_(Rune-keeper)_-_Tier_7"         , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 8), minLevel = 50, url = "Class_Deeds_(Rune-keeper)_-_Tier_8"         , section = "Class" };
            { name = L:GetText("RuneKeeper1")             , minLevel = 39, url = "Thunder_and_Flame"                          , section = "Class" };
            { name = L:GetText("RuneKeeper2")             , minLevel = 39, url = "Whispers_in_the_Dark"                       , section = "Class" };
            { name = L:GetText("RuneKeeper3")             , minLevel = 39, url = "Golu_o_Maeth"                               , section = "Class" };
            { name = L:GetText("MinesOfMoria")            , minLevel = 55, url = "The_Mines_of_Moria_(Trait_Point_Deed)"      , section = "Epic"  };
            { name = L:GetText("RuneKeeper4")             , minLevel = 58, url = "Quest:A_Rune_of_Restoration"                , section = "Class" };
            { name = L:GetText("RuneKeeper5")             , minLevel = 58, url = "On_the_Patterns_of_Wind_and_Rain"           , section = "Class" };
        };
        Warden = {
            { name = L:GetText("Tiered"):gsub("<tier>", 1), minLevel = 5 , url = "Class_Deeds_(Warden)_-_Tier_1"              , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 2), minLevel = 10, url = "Class_Deeds_(Warden)_-_Tier_2"              , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 3), minLevel = 15, url = "Class_Deeds_(Warden)_-_Tier_3"              , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 4), minLevel = 24, url = "Class_Deeds_(Warden)_-_Tier_4"              , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 5), minLevel = 30, url = "Class_Deeds_(Warden)_-_Tier_5"              , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 6), minLevel = 34, url = "Class_Deeds_(Warden)_-_Tier_6"              , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 7), minLevel = 40, url = "Class_Deeds_(Warden)_-_Tier_7"              , section = "Class" };
            { name = L:GetText("Tiered"):gsub("<tier>", 8), minLevel = 60, url = "Class_Deeds_(Warden)_-_Tier_8"              , section = "Class" };
            { name = L:GetText("Warden1")                 , minLevel = 39, url = "The_Watch_Against_the_Night"                , section = "Class" };
            { name = L:GetText("Warden2")                 , minLevel = 39, url = "Chieftains_of_the_Dúnedain"                 , section = "Class" };
            { name = L:GetText("Warden3")                 , minLevel = 39, url = "Bullroarer%27s_Boy"                         , section = "Class" };
            { name = L:GetText("MinesOfMoria")            , minLevel = 55, url = "The_Mines_of_Moria_(Trait_Point_Deed)"      , section = "Epic"  };
            { name = L:GetText("Warden4")                 , minLevel = 58, url = "The_Path_of_the_Masterful_Fist"             , section = "Class" };
            { name = L:GetText("Warden5")                 , minLevel = 58, url = "The_Path_Less_Trod"                         , section = "Class" };
        };
    }
    for tier, info in ipairs(classDeeds[playerClass]) do
        if (playerLevel >= info.minLevel) then
            local sectionStr = L:GetText("/ChecklistWindow/DeedLogSections/" .. info.section);
            text = L:GetText("Help"):gsub("<section>", sectionStr);
            help = CreateHelp(text, info.url, L:GetText("Height"));
            text = L:GetText("Desc"):gsub("<name>", info.name);
            AddItem(text, help, true);
        end
    end

    -- Points granted from other deeds
    L:SetContext("/ChecklistWindow/Items/OtherDeeds");
    local otherDeeds = {
        { name = L:GetText("OldAnorienQuests"), minLevel = 95, url = "Quests_of_Old_Anórien", tab = "Gondor", section = "OldAnorien" };
        { name = L:GetText("OldAnorienDeeds") , minLevel = 95, url = "Deeds_of_Old_Anórien" , tab = "Gondor", section = "OldAnorien" };
    };
    for _, info in ipairs(otherDeeds) do
        if (playerLevel >= info.minLevel) then
            local tabStr = L:GetText("/ChecklistWindow/DeedLogSections/" .. info.tab);
            local sectionStr = L:GetText("/ChecklistWindow/DeedLogSections/" .. info.section);
            text = L:GetText("Help"):gsub("<tab>", tabStr):gsub("<section>", sectionStr);
            help = CreateHelp(text, info.url, L:GetText("Height"));
            text = L:GetText("Desc"):gsub("<name>", info.name);
            AddItem(text, help, true);
        end
    end

    -- Points granted from class quests
    L:SetContext("/ChecklistWindow/Items/ClassQuests");
    if ((playerClass == "Beorning") or (playerClass == "Brawler")) then
        local classQuests;
        if (playerClass == "Beorning") then
            classQuests = {
                { name = L:GetText("Beorning/Quest1/Name") , from = L:GetText("Beorning/Quest1/Bestower"), minLevel = 15 , url = "Quest:The_Speech_of_Animals"   };
                { name = L:GetText("Beorning/Quest2/Name") , from = L:GetText("Beorning/Quest2/Bestower"), minLevel = 30 , url = "Quest:Hatred_of_Bear_and_Man"  };
                { name = L:GetText("Beorning/Quest3/Name") , from = L:GetText("Beorning/Quest3/Bestower"), minLevel = 45 , url = "Quest:Grimbeorn%27s_Challenge" };
                { name = L:GetText("Beorning/Quest4/Name") , from = L:GetText("Beorning/Quest4/Bestower"), minLevel = 58 , url = "Quest:The_Path_Homeward"       };
            };
        else -- (playerClass == "Brawler") --
            classQuests = {
                { name = L:GetText("Brawler/Quest1/Name") , from = L:GetText("Brawler/Quest1/Bestower"), minLevel = 15 , url = "Quest:Threat_of_a_Good_Punch"    };
                { name = L:GetText("Brawler/Quest2/Name") , from = L:GetText("Brawler/Quest2/Bestower"), minLevel = 30 , url = "Quest:The_Melee_in_the_Square"   };
                { name = L:GetText("Brawler/Quest3/Name") , from = L:GetText("Brawler/Quest3/Bestower"), minLevel = 45 , url = "Quest:Quest:Brawling_Lessons"    };
                { name = L:GetText("Brawler/Quest4/Name") , from = L:GetText("Brawler/Quest4/Bestower"), minLevel = 58 , url = "Quest:A_Drubbing_in_the_Deep"    };
            };
        end
        for _, info in ipairs(classQuests) do
            if (playerLevel >= info.minLevel) then
                text = L:GetText("Help"):gsub("<class>", playerClassStr);
                help = CreateHelp(text, url, L:GetText("Height"));
                text = L:GetText("Desc"):gsub("<name>", info.name) .. " <rgb=#808080>(" .. info.from .. ")</rgb>";
                AddItem(text, help, false);
            end
        end
    else
        -- Other classes have a single class quest
        if (playerLevel >= 45) then
            -- Temporarily change to English to get the English quest name to generate the URL
            local language = L:GetLanguage();
            L:SetLanguage(Turbine.Language.English);
            englishName = L:GetText(playerClass);
            L:SetLanguage(language);
            local url = "Quest:" .. englishName:gsub(" ", "_");
            text = L:GetText("Help"):gsub("<class>", playerClassStr);
            help = CreateHelp(text, url, L:GetText("Height"));
            name = "[50] " .. L:GetText(playerClass);
            text = L:GetText("Desc"):gsub("<name>", name);
            AddItem(text, help, false);
        end
    end

    -- Points granted from other quests
    L:SetContext("/ChecklistWindow/Items/OtherQuests");
    local otherQuests = {
        { name = L:GetText("Quest1/Name") , from = L:GetText("Quest1/Bestower"),  level = 88 , url = "Quest:The_Road_to_Dunharrow"                          , section = "Kingstead"          , aria = true  };
        { name = L:GetText("Quest2/Name") , from = L:GetText("Quest2/Bestower"),  level = 89 , url = "Quest:All_Roads_Lead_Back_to_Aldburg"                 , section = "Eastfold"           , aria = true  };
        { name = L:GetText("Quest3/Name") , from = L:GetText("Quest3/Bestower"),  level = 91 , url = "Quest:The_Broadacres_Betrayed"                        , section = "Broadacres"         , aria = false };
        { name = L:GetText("Quest4/Name") , from = L:GetText("Quest4/Bestower"),  level = 93 , url = "Quest:Woodhurst_Has_Fallen"                           , section = "Stonedeans"         , aria = false };
        { name = L:GetText("Quest5/Name") , from = L:GetText("Quest5/Bestower"),  level = 94 , url = "Quest:To_Helm%27s_Deep"                               , section = "Westfold"           , aria = true  };
        { name = L:GetText("Quest6/Name") , from = L:GetText("Quest6/Bestower"),  level = 95 , url = "Quest:A_Ruthless_End"                                 , section = "RingloVale"         , aria = true  };
        { name = L:GetText("Quest7/Name") , from = L:GetText("Quest7/Bestower"),  level = 100, url = "Quest:Blood_for_Blood"                                , section = "DorEnErnil"         , aria = false };
        { name = L:GetText("Quest8/Name") , from = L:GetText("Quest8/Bestower"),  level = 95,  url = "Quest:Sons_of_the_Usurper"                            , section = "LowerLebennin"      , aria = true  };
        { name = L:GetText("Quest9/Name") , from = L:GetText("Quest9/Bestower"),  level = 95,  url = "Quest:Faltharan%27s_Confrontation"                    , section = "Pelargir"           , aria = true  };
        { name = L:GetText("Quest10/Name"), from = L:GetText("Quest10/Bestower"), level = 95 , url = "Quest:Ashes_and_Stars,_Chapter_4"                     , section = "AshesAndStars"      , aria = true  };
        { name = L:GetText("Quest11/Name"), from = L:GetText("Quest11/Bestower"), level = 100, url = "Quest:Book_2:_Chapter_9:_Even_In_Darkness"            , section = "TheDawnlessDay"     , aria = false };
        { name = L:GetText("Quest12/Name"), from = L:GetText("Quest12/Bestower"), level = 100, url = "Quest:Book_4,_Chapter_10:_The_Defence_of_Minas_Tirith", section = "SiegeOfTheWhiteCity", aria = false };
        { name = L:GetText("Quest13/Name"), from = L:GetText("Quest13/Bestower"), level = 100, url = "Quest:Book_4,_Chapter_11:_Hammer_of_the_Underworld"   , section = "SiegeOfTheWhiteCity", aria = false };
        { name = L:GetText("Quest14/Name"), from = L:GetText("Quest14/Bestower"), level = 105, url = "Quest:Book_8,_Chapter_7:_Mordor_Triumphant"           , section = "TheBlackGate"       , aria = false };
        { name = L:GetText("Quest15/Name"), from = L:GetText("Quest15/Bestower"), level = 105, url = "Quest:Book_9,_Chapter_5:_The_Next_Adventure"          , section = "TheRingOfPower"     , aria = false };
        { name = L:GetText("Quest16/Name"), from = L:GetText("Quest16/Bestower"), level = 115, url = "Quest:Chapter_4.4:_Union_of_Evil"                     , section = "BlackBookOfMordor"  , aria = false };
        { name = L:GetText("Quest17/Name"), from = L:GetText("Quest17/Bestower"), level = 115, url = "Quest:Chapter_5.5:_The_Walls_Brought_Down"            , section = "BlackBookOfMordor"  , aria = false };
        { name = L:GetText("Quest18/Name"), from = L:GetText("Quest18/Bestower"), level = 120, url = "Quest:Chapter_8.7:_The_First_Promise"                 , section = "BlackBookOfMordor"  , aria = false };
        { name = L:GetText("Quest19/Name"), from = L:GetText("Quest19/Bestower"), level = 120, url = "Quest:Chapter_10.7:_The_Arrival_of_the_Wise"          , section = "BlackBookOfMordor"  , aria = false };
        { name = L:GetText("Quest20/Name"), from = L:GetText("Quest20/Bestower"), level = 122, url = "Quest:Chapter_12.5:_The_End_of_the_Tale"              , section = "BlackBookOfMordor"  , aria = false };
        { name = L:GetText("Quest21/Name"), from = L:GetText("Quest21/Bestower"), level = 128, url = "Quest:Chapter_14.5:_A_Final_Escape"                   , section = "BlackBookOfMordor"  , aria = false };
        { name = L:GetText("Quest22/Name"), from = L:GetText("Quest22/Bestower"), level = 130, url = "Quest:Chapter_2.7:_A_Thirst_For_Blood"                , section = "LegacyOfDurin"      , aria = false };
        { name = L:GetText("Quest23/Name"), from = L:GetText("Quest23/Bestower"), level = 1,   url = "Quest:Book_1,_Chapter_8:_The_Wedding_Banquet"         , section = "ATimeOfCelebration" , aria = false };
        { name = L:GetText("Quest24/Name"), from = L:GetText("Quest24/Bestower"), level = 130, url = "Quest:Chapter_4.9:_The_Lord_of_Gundabad"              , section = "LegacyOfDurin"      , aria = false };
        { name = L:GetText("Quest25/Name"), from = L:GetText("Quest25/Bestower"), level = 130, url = "Quest:Interlude:_Blood_of_Azog"                       , section = "BloodOfAzog"        , aria = false };
        { name = L:GetText("Quest26/Name"), from = L:GetText("Quest26/Bestower"), level = 136, url = "Quest:Chapter_7.8:_A_Silent_Presence"                 , section = "LegacyOfDurin"      , aria = false };
        { name = L:GetText("Quest27/Name"), from = L:GetText("Quest27/Bestower"), level = 140, url = "Quest:Chapter_11.5:_Gundabad_Is_Ours!"                , section = "LegacyOfDurin"      , aria = false };
    };
    for _, info in ipairs(otherQuests) do
        if (playerLevel >= info.level) then
            local sectionStr = L:GetText("/ChecklistWindow/QuestLogSections/" .. info.section);
            text = L:GetText("Help"):gsub("<section>", sectionStr);
            help = CreateHelp(text, info.url, L:GetText("Height"));
            name = "[" .. info.level .. "] " .. info.name;
            text = L:GetText("Desc"):gsub("<name>", name) .. " <rgb=#808080>(" .. info.from .. ")</rgb>";
            AddItem(text, help, info.aria);
        end
    end

    -- Points granted for Epic Battles
    L:SetContext("/ChecklistWindow/Items/EpicBattles");
    if (playerLevel >= 10) then
        for deed = 1, 2 do
            local url = "Promotion_Points_" .. deed;
            text = L:GetText("Help"):gsub("<points>", deed * 100):gsub("<deed>", deed);
            help = CreateHelp(text, url, L:GetText("Height"));
            text = L:GetText("Desc"):gsub("<points>", deed * 100);
            AddItem(text, help, true);
        end
    end

    self:RecalculateTotal();
    DoCallbacks(self, "SizeChanged");
end

function Checklist:PositionChanged()
    local left, top = self:GetPosition();
    self.settings.position = { left, top };
    self:SaveSettings();
end

function Checklist:_SizeChanged()
    if (not self.loading) then
        local width, height = self:GetSize();
        self.settings.size = { width, height };
        self:SaveSettings();
    end
end

function Checklist:RecalculateTotal()
    local total, possible = 0, 0;
    for _, checkBox in pairs(self.checkBoxes) do
        if (checkBox.visible) then
            possible = possible + 1;
            if (checkBox:IsChecked()) then
                total = total + 1;
            end
        end
    end
    if (self.ariaCheckBox) then
        possible = possible - 1;
        if (self.ariaCheckBox:IsChecked()) then
            total = total - 1 + self.ariaPoints;
            possible = possible + self.ariaPoints;
        end
    end
    self.total:SetText(total .. " / " .. possible);
    self:SaveSettings();
end

function Checklist:SaveSettings()
    -- Only do a save at logout.  This plugin isn't so important that we need to be more careful.
end

function Checklist:ChangeLanguage(newLanguage)
    Turbine.UI.Lotro.Window.SetVisible(self, false);
    Turbine.UI.Lotro.Window.SetVisible(self.icon, false);
    self.settings.language = newLanguage;
    L:SetLanguage(newLanguage);
    SetCyrillicEnabled(newLanguage == Turbine.Language.Russian);
    -- Wait until next frame to create new Checklist instance and assign it to _G.win.
    -- This is just in case we are in the process of loading the plugin.
    self:SetWantsUpdates(true);
    -- continnues in CheckList:Update()
end

function Checklist:Update()
    -- continuing from CheckList:ChangeLanguage():
    self:SetWantsUpdates(false);
    self:Close();
    _G.win = Checklist(self.settings, self.orderedCheckBoxes, self.orderedContainers);
end

function Checklist:Close()
    Turbine.PluginData.Save(Turbine.DataScope.Character, "ClassTraitPoints", ExportTable(self.settings));
    Turbine.UI.Lotro.Window.Close(self);
end

function Checklist:SetVisible(visible)
    self.settings.visible = visible;
    Turbine.UI.Lotro.Window.SetVisible(self, visible);
    self:SaveSettings();
end

function Checklist:Closing(args)
    self.icon:Iconify();
    args.Cancel = true;
end
